#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/time.h>
#include <linux/random.h>
#include <linux/sched.h>

#include <linux/tcfs.h>
#include <linux/tcfs_fs.h>
#include <linux/tcfs/tcfs_hash.h>
#include <linux/tcfs/cipherinterface.h>

#include <linux/tcfs/global.h>
#include <linux/tcfs/md5.h>

#define TCFS_DEBUG	(TCFS_DBG_NONE)
/* One of:
#define TCFS_DEBUG	(TCFS_DBG_GEN|TCFS_DBG_FUNC_IN|TCFS_DBG_FUNC_OUT|TCFS_DBG_ERR)
*/

extern u_int tcfs_numreadflags, tcfs_numreadget, tcfs_numofgetheader;
extern u_int tcfs_numofputheader;
extern u_int tcfs_numputwrite, tcfs_numputflags;

static
void tcfs_printk (unsigned int facility, char *msg)
{
	switch (facility&TCFS_DEBUG) {
		case TCFS_DBG_NONE:
			return ;
		case TCFS_DBG_FUNC_IN:
			printk ("TCFS: in -");
			break;
		case TCFS_DBG_FUNC_OUT:
			printk ("TCFS: out -");
			break;
		case TCFS_DBG_GEN:
			printk ("TCFS:");
			break;
		case TCFS_DBG_ERR:
			printk ("TCFS: error ->");
			break;
	}
	printk (" %s\n", msg);
}

extern
int tcfs_instantiate(struct dentry *, struct tcfs_fh *, struct tcfs_fattr *);

static unsigned int
tcfs_ldiv (long long opa, long long opb)
{
	unsigned int res=0;

	while (( opa -= opb) >= 0) {
		res++;
	}

	return res;
}

static void
tcfs_hash_block (const char *buffer, int size, char *out)
{
	MD5_CTX context;
	unsigned char digest[TCFS_HASH_SIZE];

	memset (digest, 0, TCFS_HASH_SIZE);

	MD5Init (&context);
	MD5Update (&context, buffer, size);
	MD5Final (digest, &context);

        memcpy (out, digest, TCFS_HASH_SIZE);
}               
        
static void 
tcfs_get_fk (void *fk)
{
	get_random_bytes (fk, KEYSIZE);
}

int
tcfs_check_hash (const char *buffer, const char *bk, char *hash, int size)
{ 
	char *newhash=NULL;
	char buf[size+KEYSIZE];

	newhash=kmalloc (TCFS_HASH_SIZE, GFP_KERNEL);
	if (!newhash)
		return -ENOMEM;
	
	memcpy (buf, buffer, size);

	if (bk) {
		memcpy (buf+size, bk, KEYSIZE);
		tcfs_hash_block (buf, size+KEYSIZE, newhash);
	} else {
		tcfs_hash_block (buf, size, newhash);
	}

	if (!memcmp (hash, newhash, TCFS_HASH_SIZE)) {
		kfree (newhash);

		return -TCFS_ERROK;
	} else {
		kfree (newhash);

		return -TCFS_ERRGEN;
	}
}               

int
tcfs_get_header (struct inode *inode, struct tcfs_header *tcfsheader, void *key)
{
        int res;
        char *buffer=NULL;
        struct tcfs_fattr fattr;
	struct tcfs_fh fhandle;
	int error=-TCFS_ERROK;

	tcfs_numofgetheader++;

	if (!inode->u.tcfs_i.fhandle) {
		return -TCFS_ERRGEN;
	}
	
	if (inode->u.tcfs_i.cached==2) {
		memcpy (tcfsheader->start, "TCFS", 4);
		memcpy (tcfsheader->engine, inode->u.tcfs_i.tcfsheader.engine,
			TCFS_ENGINE_FIELD_SIZE);
		memcpy (tcfsheader->file_key,
			inode->u.tcfs_i.tcfsheader.file_key, KEYSIZE);
		tcfsheader->size = inode->u.tcfs_i.tcfsheader.size;
		tcfsheader->flags = inode->u.tcfs_i.tcfsheader.flags;

		return -TCFS_ERROK;
	}

	if (inode->u.tcfs_i.cached==1 && !key) {
		memcpy (tcfsheader->start, "TCFS", 4);
		memcpy (tcfsheader->engine, inode->u.tcfs_i.tcfsheader.engine,
			TCFS_ENGINE_FIELD_SIZE);
		tcfsheader->size = inode->u.tcfs_i.tcfsheader.size;
		tcfsheader->flags = inode->u.tcfs_i.tcfsheader.flags;

		return -TCFS_ERROK;
	}

	if (inode->u.tcfs_i.cached==0) { 
		buffer=kmalloc (TCFS_HEADER_SIZE+TCFS_HASH_SIZE, GFP_KERNEL);
		if (!buffer) {
			return -ENOMEM;
		}

		inode->u.tcfs_i.tcfsheader.flags = 0;
		memset (buffer, 0, TCFS_HEADER_SIZE+TCFS_HASH_SIZE);

		if (S_ISDIR(inode->i_mode)) {
			error=tcfs_proc_lookup (TCFS_SERVER(inode),
				TCFS_INODE_FH(inode), ".tcfs",
				&fhandle, &fattr);
			if (error<0) {
				return -TCFS_ERRGEN;
			}
		} else {
			fhandle = *(struct tcfs_fh *)TCFS_INODE_FH(inode);
		}
	
		tcfs_numreadget++;
		res=tcfs_proc_read (TCFS_SERVER(inode), &fhandle, 0, 0,
                	TCFS_HEADER_SIZE+TCFS_HASH_SIZE, buffer, &fattr);

		if (res<0) {
			kfree (buffer);

			return res;
		}

		if (res==0) {
			kfree (buffer);

			return -TCFS_ERRGEN;
		}

		if (res!=TCFS_HEADER_SIZE+TCFS_HASH_SIZE) {
			kfree (buffer);

			return -TCFS_ERRGEN;
		}
	
		error=tcfs_check_hash (buffer, NULL, buffer+TCFS_HEADER_SIZE,
			 TCFS_HEADER_SIZE);
		if (error<0) {
			kfree (buffer);

			return -TCFS_ERRGEN;
		}

        	if (memcmp (((struct tcfs_header *)buffer)->start,"TCFS",4)) {
			kfree (buffer);

			return -TCFS_ERRGEN;
		}

		memcpy (tcfsheader, buffer, sizeof(struct tcfs_header));
		kfree (buffer);

		inode->u.tcfs_i.cached=1;
		inode->u.tcfs_i.tcfsheader.size = tcfsheader->size;
		inode->u.tcfs_i.tcfsheader.flags = tcfsheader->flags;
		memcpy (inode->u.tcfs_i.tcfsheader.engine, tcfsheader->engine,
			TCFS_ENGINE_FIELD_SIZE);
		memcpy (inode->u.tcfs_i.tcfsheader.file_key,
			tcfsheader->file_key, KEYSIZE);

		if (key) {
			error=tcfs_decrypt (tcfsheader->file_key, KEYSIZE, key,
				tcfsheader->engine);
			if (error<0) {
				return error;
			}

			memcpy (inode->u.tcfs_i.tcfsheader.file_key,
				tcfsheader->file_key, KEYSIZE);
			inode->u.tcfs_i.cached=2;
		}

		return -TCFS_ERROK;
	}

	if (inode->u.tcfs_i.cached==1) {
		memcpy (tcfsheader->start, "TCFS", 4);
		memcpy (tcfsheader->engine, inode->u.tcfs_i.tcfsheader.engine,
			TCFS_ENGINE_FIELD_SIZE);
		error=tcfs_decrypt (inode->u.tcfs_i.tcfsheader.file_key,
				KEYSIZE, key, tcfsheader->engine);
		if (error<0) {
			inode->u.tcfs_i.cached=0;

			return error;
		}
		memcpy (tcfsheader->file_key,
			inode->u.tcfs_i.tcfsheader.file_key, KEYSIZE);
		tcfsheader->size = inode->u.tcfs_i.tcfsheader.size;
		tcfsheader->flags = inode->u.tcfs_i.tcfsheader.flags;
		inode->u.tcfs_i.cached=2;
	}

	return -TCFS_ERROK;
}

static int
tcfs_get_flag (struct inode *inode)
{
	struct tcfs_header tcfsheader;

	if (!inode->u.tcfs_i.cached) {
		int error=-TCFS_ERROK;

		error=tcfs_get_header (inode, &tcfsheader, NULL);
		if (error<0) {
			return error;
		}
	}

	if (TCFS_IS_SECURE(inode)) {
		return TCFS_SECURE_FL;
	}

	if (TCFS_IS_SHARED(inode)) {
		return TCFS_SHARED_FL;
	}

	return TCFS_CLEAR_FILE;
}

void *
tcfs_get_key (struct inode *inode)
{
	struct user_key *uk=NULL;
	struct group_key *gk=NULL;
	char *engine;

	switch (tcfs_get_flag (inode)) {
		case TCFS_SECURE_FL:
			engine=inode->u.tcfs_i.tcfsheader.engine;
			uk=tcfs_key_lookup (current->uid, engine);
			if (!uk) {
				return NULL;
			}

			if (inode->i_uid==current->uid) {
				return uk->key;
			}
	
			return NULL;

		case TCFS_SHARED_FL:
			engine=inode->u.tcfs_i.tcfsheader.engine;
			gk=tcfs_groupkey_lookup (inode->i_gid, engine);
			if (!gk || !gk->key) {
				return NULL;
			}

			if (inode->i_gid == current->gid)
				return gk->key;

			return NULL;
	}

	return TCFS_NO_SECURE;
}

int
tcfs_put_header (struct file *filp, struct dentry *dentry, struct tcfs_header *tcfsheader, void *key)
{
        char *buffer=NULL, *hres=NULL;
	struct tcfs_fh *fhandle=NULL;
	struct inode *inode=NULL, *inode1=NULL;
	struct dentry *dentry1=NULL;
	struct tcfs_fh fhandle1;
        struct tcfs_fattr fattr;
        int error=0;
	char *engine=NULL;

	tcfs_numofputheader++;

	engine=tcfsheader->engine;

	if (!dentry && filp) {
		dentry=filp->f_dentry;
		inode=dentry->d_inode;
	} else {
		inode=dentry->d_inode;
	}

	fhandle=(struct tcfs_fh *)dentry->d_fsdata;

	if (S_ISDIR(inode->i_mode)) {
		struct tcfs_sattr sattr;
		struct qstr qname;
		const char *name=".tcfs";

		sattr.mode = S_IRUSR|S_IWUSR;
		sattr.uid = sattr.gid = sattr.size = (unsigned) -1;
		sattr.atime.seconds = sattr.mtime.seconds = (unsigned) -1;

		tcfs_invalidate_dircache(dentry->d_parent->d_inode);

		qname.name=name;
		qname.len=5;
		
		dentry1=d_alloc (dentry, &qname);
		if (!dentry1) {
			return -ENOMEM;
		}
		
		dentry1->d_fsdata=kmalloc(sizeof(struct tcfs_fh), GFP_KERNEL);
		if (!dentry1->d_fsdata) {
			return -ENOMEM;
		}

		error=tcfs_proc_create (TCFS_SERVER(inode), fhandle, name,
			&sattr, &fhandle1, &fattr);

		if (!error) {
			error=tcfs_instantiate(dentry1, &fhandle1, &fattr);
		}

		if (error) {
			d_drop (dentry1);
			dput (dentry1);

			return error;
		}
		
		inode1=dentry1->d_inode;
	} else {
		inode1=inode;
		fhandle1=*fhandle;
	}

        buffer=kmalloc (TCFS_HEADER_SIZE+TCFS_HASH_SIZE, GFP_KERNEL);
	if (!buffer) {
		if (dentry1) {
			d_drop (dentry1);
			dput (dentry1);
		}

		return -ENOMEM;
	}

	hres=kmalloc (TCFS_HASH_SIZE, GFP_KERNEL);
	if (!hres) {
		kfree (buffer);
		if (dentry1) {
			d_drop (dentry1);
			dput (dentry1);
		}

		return -ENOMEM;
	}
	
	error=tcfs_encrypt (tcfsheader->file_key, KEYSIZE, key, engine);
	if (error<0) {
		kfree (buffer);
		kfree (hres);

		return error;
	}

	memset (buffer, 0, TCFS_HEADER_SIZE+TCFS_HASH_SIZE);
        memcpy (buffer, tcfsheader, sizeof(struct tcfs_header));
	tcfs_hash_block (buffer, TCFS_HEADER_SIZE, hres);
	memcpy (buffer+TCFS_HEADER_SIZE, hres, TCFS_HASH_SIZE);
	tcfs_numputwrite++;
        error=tcfs_proc_write (TCFS_SERVER(inode), &fhandle1, 0, 0,
		TCFS_HEADER_SIZE+TCFS_HASH_SIZE, buffer, &fattr);

        kfree (buffer);
	kfree (hres);

	if (dentry1) {
		d_drop (dentry1);
		dput (dentry1);
	}

	if (error<0) {
		return error;
	} 

        return -TCFS_ERROK;
}

static int
tcfs_decrypt_file(struct dentry *dentry, struct tcfs_header *tcfsheader)
{
	int error=-TCFS_ERROK;
	char buffer[TCFS_HBLOCK_SIZE];
	char *engine=NULL;
	off_t size;
	unsigned int towrite, blocks, block_number;
	loff_t offset, newoffset;
	struct tcfs_fattr fattr;
	struct tcfs_sattr sattr;
	struct inode *inode=NULL;
	char bkey[TCFS_HASH_SIZE], buf[KEYSIZE+sizeof(unsigned int)];
	void *fk=tcfsheader->file_key;
	void *bk=NULL;

	memset (bkey, 0, TCFS_HASH_SIZE);
	memset (buf, 0, KEYSIZE+sizeof(unsigned int));
	memcpy (buf, fk, KEYSIZE);

	inode=dentry->d_inode;
	engine=tcfsheader->engine;
	size=tcfsheader->size;
	if (!size) {
		goto out;
	}

	blocks=size/TCFS_BLOCK_SIZE;
	offset=0;
	towrite=TCFS_BLOCK_SIZE;
	newoffset=TCFS_HEADER_SIZE+TCFS_HASH_SIZE;
	blocks++;
	block_number=1;

	do {
		memcpy (buf+KEYSIZE,(char *)&block_number,sizeof(unsigned int));
		tcfs_hash_block (buf, KEYSIZE+sizeof(unsigned int), bkey);
		bk=tcfs_init_key(bkey, engine);
		if (!bk) {
			return -TCFS_ERRGEN;
		}

		error=tcfs_proc_read (TCFS_SERVER(inode), TCFS_INODE_FH(inode),
			0, newoffset, TCFS_HBLOCK_SIZE, buffer, &fattr);
		if (error<0) {
			tcfs_cleanup_key (bk, engine);

			return error;
		}

		error=tcfs_decrypt (buffer, TCFS_HBLOCK_SIZE, bk, engine);
		if (error<0) {
			tcfs_cleanup_key (bk, engine);

			return error;
		}

		error=tcfs_check_hash (buffer, bk, buffer+TCFS_BLOCK_SIZE,
			TCFS_BLOCK_SIZE);
		if (error<0) {
			tcfs_cleanup_key (bk, engine);

			return -TCFS_ERRGEN;
		}

		error=tcfs_proc_write (TCFS_SERVER(inode), TCFS_INODE_FH(inode),
			0, offset, towrite, buffer, &fattr);
		if (error<0) {
			return error;
		}

		error=tcfs_cleanup_key (bk, engine);
		if (error<0) {
			return error;
		}

		block_number++;
		offset+=TCFS_BLOCK_SIZE;
		newoffset+=TCFS_HBLOCK_SIZE;
		towrite=TCFS_BLOCK_SIZE;
	} while (block_number<=blocks);

out:
	sattr.mode=inode->i_mode;
	sattr.size=inode->i_size;
	sattr.uid=inode->i_uid;
	sattr.gid=inode->i_gid;

	error=tcfs_proc_setattr(TCFS_DSERVER(dentry), TCFS_FH(dentry),
		&sattr, &fattr);
	if (error) {
		return error;
	}

	return -TCFS_ERROK;
}

static int
tcfs_encrypt_file(struct inode *inode, struct tcfs_header *tcfsheader)
{
	int error=-TCFS_ERROK;
	char buffer[TCFS_HBLOCK_SIZE], hash[TCFS_HASH_SIZE];
	char *engine=NULL;
	struct tcfs_fattr fattr;
	off_t size;
	unsigned int toread, blocks;
	loff_t offset, newoffset;
	char bkey[TCFS_HASH_SIZE], buf[KEYSIZE+sizeof(unsigned int)];
	void *fk=tcfsheader->file_key;
	void *bk=NULL;

	memset (bkey, 0, TCFS_HASH_SIZE);
	memset (buf, 0, KEYSIZE+sizeof(unsigned int));
	memset (buffer, 0, TCFS_HBLOCK_SIZE);
	memcpy (buf, fk, KEYSIZE);

	engine=tcfsheader->engine;
	size=tcfsheader->size;
	if (!size) {
		return -TCFS_ERROK;
	}

	blocks=size/TCFS_BLOCK_SIZE;
	offset=blocks*TCFS_BLOCK_SIZE;
	toread=size-offset; /* last block */
	newoffset=blocks*TCFS_HBLOCK_SIZE+TCFS_HEADER_SIZE+TCFS_HASH_SIZE;
	blocks++;

	do {
		memcpy (buf+KEYSIZE, (char *)&blocks, sizeof(unsigned int));
		tcfs_hash_block (buf, KEYSIZE+sizeof(unsigned int), bkey);
		bk=tcfs_init_key (bkey, engine);
		if (!bk) {
			return -TCFS_ERRGEN;
		}

		error=tcfs_proc_read (TCFS_SERVER(inode), TCFS_INODE_FH(inode),
			0, offset, toread, buffer, &fattr);
		if (error<0) {
			tcfs_cleanup_key (bk, engine);

			return error;
		}

		memcpy (buffer+TCFS_BLOCK_SIZE, bk, KEYSIZE);
		tcfs_hash_block (buffer, TCFS_BLOCK_SIZE+KEYSIZE, hash);
		memcpy (buffer+TCFS_BLOCK_SIZE, hash, TCFS_HASH_SIZE);
		error=tcfs_encrypt (buffer, TCFS_HBLOCK_SIZE, bk, engine);
		if (error<0) {
			tcfs_cleanup_key (bk, engine);

			return error;
		}

		error=tcfs_proc_write (TCFS_SERVER(inode), TCFS_INODE_FH(inode),
			0, newoffset, TCFS_HBLOCK_SIZE, buffer, &fattr);
		if (error<0) {
			tcfs_cleanup_key (bk, engine);

			return error;
		}

		error=tcfs_cleanup_key (bk, engine);
		if (error<0) {
			return error;
		}

		blocks--;
		offset-=TCFS_BLOCK_SIZE;
		newoffset-=TCFS_HBLOCK_SIZE;
		toread=TCFS_BLOCK_SIZE;
	} while (blocks);

	return -TCFS_ERROK;
}

static int
tcfs_reencrypt_file (struct inode *inode, char *newcipher, struct tcfs_header *tcfsheader)
{
	int error=-TCFS_ERROK;
	char buffer[TCFS_HBLOCK_SIZE], hash[TCFS_HASH_SIZE];
	char *oldcipher=NULL;
	struct tcfs_fattr fattr;
	off_t size;
	unsigned int blocks, offset, block_number;
	char buf[KEYSIZE+sizeof (unsigned int)];
	char bkey[TCFS_HASH_SIZE];
	void *fk=tcfsheader->file_key;
	void *bk=NULL;

	memset (bkey, 0, TCFS_HASH_SIZE);
	memset (buf, 0, KEYSIZE+sizeof(unsigned int));
	memcpy (buf, fk, KEYSIZE);

	oldcipher=tcfsheader->engine;
	size=tcfsheader->size;
	if (!size) {
		return -TCFS_ERROK;
	}

	blocks=size/TCFS_BLOCK_SIZE;
	offset=TCFS_HEADER_SIZE+TCFS_HASH_SIZE; /* Start of the file */

	blocks++;
	block_number=1;

	do {
		memcpy (buf+KEYSIZE, (char *)&block_number, sizeof(unsigned int));
		tcfs_hash_block (buf, KEYSIZE+sizeof(unsigned int), bkey);
		bk=tcfs_init_key (bkey, oldcipher);
		if (!bk) {
			return -TCFS_ERRGEN;
		}

		error=tcfs_proc_read (TCFS_SERVER(inode), TCFS_INODE_FH(inode),
			0, offset, TCFS_HBLOCK_SIZE, buffer, &fattr);
		if (error<0) {
			tcfs_cleanup_key (bk, oldcipher);

			return error;
		}

		error=tcfs_decrypt (buffer, TCFS_HBLOCK_SIZE, bk, oldcipher);
		if (error<0) {
			tcfs_cleanup_key (bk, oldcipher);

			return error;
		}

		error=tcfs_check_hash (buffer, bk, buffer+TCFS_BLOCK_SIZE,
				TCFS_BLOCK_SIZE);
		if (error<0) {
			tcfs_cleanup_key (bk, oldcipher);

			return error;
		}

		error=tcfs_cleanup_key (bk, oldcipher);
		if (error<0) {
			return error;
		}

		bk=tcfs_init_key (bkey, newcipher);
		if (!bk) {
			return -TCFS_ERRGEN;
		}

		memcpy (buffer+TCFS_BLOCK_SIZE, bk, KEYSIZE);
		tcfs_hash_block (buffer, TCFS_BLOCK_SIZE+KEYSIZE, hash);
		memcpy (buffer+TCFS_BLOCK_SIZE, hash, TCFS_HASH_SIZE);
		error=tcfs_encrypt (buffer, TCFS_HBLOCK_SIZE, bk, newcipher);
		if (error<0) {
			tcfs_cleanup_key (bk, newcipher);

			return error;
		}

		error=tcfs_cleanup_key (bk, newcipher);
		if (error<0) {
			return error;
		}

		error=tcfs_proc_write (TCFS_SERVER(inode), TCFS_INODE_FH(inode),
			0, offset, TCFS_HBLOCK_SIZE, buffer, &fattr);
		if (error<0) {
			return error;
		}

		block_number++;
		offset+=TCFS_HBLOCK_SIZE;
	} while (block_number <= blocks);

	return -TCFS_ERROK;
}

static int
tcfs_decrypt_dir (struct file *filp, struct tcfs_header *tcfsheader, void *key)
{
	__u32 *entry=NULL;
	char *engine=NULL, *start;
	struct dentry *dentry=filp->f_dentry;
	struct inode *dir=dentry->d_inode;
	unsigned int size, error=0, index=0;
	u32 cookie;

	entry=(__u32 *)get_free_page(GFP_KERNEL);
	if (!entry) {
		return -ENOMEM;
	}

	cookie=filp->f_pos;
	engine=tcfsheader->engine;

	error=tcfs_proc_readdir (TCFS_SERVER(dir), TCFS_FH(dentry),
		cookie, PAGE_SIZE, entry);
	if (error<0) {
		free_page ((unsigned long)entry);

		return -error;
	}

	size=error;
	start= (char *)entry;
	entry++;

	while (index<size) {
		__u32	nextpos=*entry++;
		__u32	length=*entry++;
		unsigned int len=length&0x7FFF;
		char	*name=NULL;
		char	tcfsname[TCFS_MAXNAMLEN];
		char	oldname[TCFS_MAXNAMLEN];

		name=start+((length>>16)&0x7FFF);

		memset (tcfsname, 0, TCFS_MAXNAMLEN);
		memset (oldname, 0, TCFS_MAXNAMLEN);

		memcpy (tcfsname, name, len);
		memcpy (oldname, name, len);
		tcfsname[len]=0;
		oldname[len]=0;

		if (strcmp (tcfsname, ".") && strcmp (tcfsname, "..") &&
			strcmp (tcfsname, ".tcfs")) {
			len=tcfsdecode (tcfsname, len+1);
			len=strlen(tcfsname);
			error=tcfs_decrypt (tcfsname,(len&0xFFF8)+8,key,engine);
			if (error<0) {
				return error;
			}
			
			tcfs_invalidate_dircache (dir);

			error=tcfs_proc_rename (TCFS_DSERVER(dentry), 
				TCFS_FH(dentry), oldname,
				TCFS_FH(dentry), tcfsname);
			if (error) {
				return error;
			}
		}
		cookie = nextpos;
		index++;
		entry++;
	}

	return -TCFS_ERROK;
}

static int
tcfs_encrypt_dir (struct file *filp, struct tcfs_header *tcfsheader, void *key)
{
	__u32 *entry=NULL;
	char *engine=NULL, *start=NULL;
	struct dentry *dentry=filp->f_dentry;
	struct inode *dir=dentry->d_inode;
	unsigned int size, error=0, index=0;
	u32 cookie;

	entry=(__u32 *)get_free_page(GFP_KERNEL);
	if (!entry) {
		return -ENOMEM;
	}

	cookie=filp->f_pos;
	engine=tcfsheader->engine;

	error=tcfs_proc_readdir (TCFS_SERVER(dir), TCFS_FH(dentry),
		cookie, PAGE_SIZE, entry);
	if (error<0) {
		free_page ((unsigned long)entry);
		return error;
	}

	size=error;
	start=(char *)entry;
	entry++;

	while (index<size) {
		__u32	nextpos=*entry++;
		__u32	length=*entry++;
		unsigned int len=length&0x7FFF;
		char	*name=NULL;
		char	tcfsname[TCFS_MAXNAMLEN];
		char	oldname[TCFS_MAXNAMLEN];

		name=start+((length>>16)&0x7FFF);

		memset (tcfsname, 0, TCFS_MAXNAMLEN);
		memset (oldname, 0, TCFS_MAXNAMLEN);
		memcpy (tcfsname, name, len);
		memcpy (oldname, name, len);
		tcfsname[len]=0;
		oldname[len]=0;

		if (strcmp (tcfsname, ".") && strcmp (tcfsname, "..")) {
			error=tcfs_encrypt (tcfsname, ((len+1)&0xFFF8)+8,
				key, engine);
			if (error<0) {
				return error;
			}

			tcfsencode (tcfsname, ((len+1)&0xFFF8)+8);
			tcfs_invalidate_dircache (dir);
			error=tcfs_proc_rename (TCFS_DSERVER(dentry), 
				TCFS_FH(dentry), oldname,
				TCFS_FH(dentry), tcfsname);
			if (error) {
				return error;
			}
		}
		cookie = nextpos;
		index++;
		entry++;
	}

	return -TCFS_ERROK;
}

static int
tcfs_reencrypt_dir (struct file *filp, char *newcipher, struct tcfs_header *tcfsheader, void *oldkey, void *newkey)
{
	__u32 *entry=NULL;
	char *engine=NULL, *start=NULL;
	struct dentry *dentry=filp->f_dentry;
	struct inode *dir=dentry->d_inode;
	unsigned int size, error=0, index=0;
	u32 cookie;

	entry=(__u32 *)get_free_page(GFP_KERNEL);
	if (!entry) {
		return -ENOMEM;
	}

	cookie=filp->f_pos;
	engine=tcfsheader->engine;

	error=tcfs_proc_readdir (TCFS_SERVER(dir), TCFS_FH(dentry),
		cookie, PAGE_SIZE, entry);
	if (error<0) {
		free_page ((unsigned long)entry);

		return -TCFS_ERRGEN;
	}

	size=error;
	start=(char *)entry;
	entry++;

	while (index<size) {
		__u32	nextpos=*entry++;
		__u32	length=*entry++;
		unsigned int len=length&0x7FFF;
		char	*name=NULL;
		char	tcfsname[TCFS_MAXNAMLEN];
		char	oldname[TCFS_MAXNAMLEN];

		name=start+((length>>16)&0x7FFF);

		memset (tcfsname, 0, TCFS_MAXNAMLEN);
		memset (oldname, 0, TCFS_MAXNAMLEN);
		memcpy (tcfsname, name, len);
		memcpy (tcfsname, name, len);
		tcfsname[len]=0;
		oldname[len]=0;

		if (strcmp (tcfsname, ".") && strcmp (tcfsname, "..")
			&& strcmp (tcfsname, ".tcfs")) {
			len=tcfsdecode (tcfsname, len+1);
			len=strlen(tcfsname);
			error=tcfs_decrypt (tcfsname, ((len+1)&0xFFF8)+8,
					oldkey, engine);
			if (error<0) {
				return error;
			}
	
			error=tcfs_encrypt (tcfsname, ((len+1)&0xFFF8)+8,
					newkey, newcipher);
			if (error<0) {
				return error;
			}

			tcfsencode (tcfsname, ((len+1)&0xFFF8)+8);
			error=tcfs_proc_rename (TCFS_DSERVER(dentry), 
				TCFS_FH(dentry), oldname,
				TCFS_FH(dentry), tcfsname);
			if (error) {
				return error;
			}
		}

		cookie = nextpos;
		index++;
		entry++;
	}

	return -TCFS_ERROK;
}

static int
tcfs_decrypt_it (struct file *filp, void *key)
{
	int error=-TCFS_ERROK;
	struct inode *inode=filp->f_dentry->d_inode;
	struct tcfs_header tcfsheader;

	error=tcfs_get_header (inode, &tcfsheader, key);
	if (error<0) {
		return error;
	}

	if (S_ISDIR(inode->i_mode)) {
		error=tcfs_decrypt_dir (filp, &tcfsheader, key);
		if (error<0) {
			return error;
		}

		error=tcfs_proc_remove (TCFS_SERVER(inode),
			(struct tcfs_fh *)filp->f_dentry->d_fsdata,
			".tcfs");
		if (error) {
			return error;
		}

		tcfs_invalidate_dircache (inode);
	} else {
		error=tcfs_decrypt_file (filp->f_dentry, &tcfsheader);
	}

	if (error<0) {
		return error;
	}

	return -TCFS_ERROK;
}

static int
tcfs_encrypt_it (struct file *filp, void *key, char *engine, unsigned int flag)
{
	int error=-TCFS_ERROK;
	struct inode *inode=filp->f_dentry->d_inode;
	struct tcfs_header tcfsheader;
	
	memset (&tcfsheader, 0, sizeof (struct tcfs_header));

	tcfsheader.size=inode->i_size;
	memcpy (tcfsheader.engine, engine, TCFS_ENGINE_FIELD_SIZE);
	memcpy (tcfsheader.start, "TCFS", 4);
	/* version */
	tcfsheader.flags|=flag;
	tcfs_get_fk (tcfsheader.file_key);

	if (S_ISDIR(inode->i_mode)) {
		error=tcfs_encrypt_dir (filp, &tcfsheader, key);
	} else {
		error=tcfs_encrypt_file (inode, &tcfsheader);
	}

	if (error<0) {
		return error;
	}

	error=tcfs_put_header (filp, NULL, &tcfsheader, key);

	return error;
}

static int
tcfs_flags_changed (struct file *filp, struct tcfs_flags *tcfsflags, void *key)
{
	int error=0;
	unsigned int oldflags;
	unsigned int newflags=tcfsflags->flags;
	unsigned int file_is_secure, file_is_shared, want_secure, want_shared;
	char *engine=tcfsflags->engine;

	oldflags=filp->f_dentry->d_inode->u.tcfs_i.tcfsheader.flags;

	file_is_secure=oldflags&TCFS_SECURE_FL;
	file_is_shared=oldflags&TCFS_SHARED_FL;
	want_secure=newflags&TCFS_SECURE_FL;
	want_shared=newflags&TCFS_SHARED_FL;

	if (file_is_secure && file_is_shared) {
		printk ("TCFS: inode %ld has both group and secure flag!\n",
			filp->f_dentry->d_inode->i_ino);

		return -TCFS_ERRINVAL;
	}

	if (want_secure && want_shared) {
		printk ("TCFS: cannot set both group and secure flag!\n");

		return -TCFS_ERRINVAL;
	}

	if (want_secure && file_is_secure) {
		return -TCFS_ERRINVAL;
	}

	if (want_shared && file_is_shared) {
		return -TCFS_ERRINVAL;
	}

	if (want_shared && file_is_secure) {
		return -TCFS_ERRINVAL;
	}

	if (want_secure && file_is_shared) {
		return -TCFS_ERRINVAL;
	}

	if (!file_is_secure && !file_is_shared) {
		if (want_secure) {
			error=tcfs_encrypt_it (filp, key, engine,
						TCFS_SECURE_FL);
		}
		
		if (want_shared) {
			error=tcfs_encrypt_it (filp, key, engine,
						TCFS_SHARED_FL);
		}
	}

	if ((file_is_shared && !want_shared) ||
			 (file_is_secure && !want_secure)) {
		error=tcfs_decrypt_it (filp, key);
	}

	if (error<0) {
		return error;
	}

	filp->f_dentry->d_inode->u.tcfs_i.tcfsheader.flags=newflags;

	return -TCFS_ERROK;
}

static int
tcfs_engine_changed (struct file *filp, char *newcipher, void *oldkey)
{
	int error=-TCFS_ERROK;
	struct tcfs_header tcfsheader;
	struct inode *inode=filp->f_dentry->d_inode;
	struct user_key *uk=NULL;
	void *newkey=NULL;

	error=tcfs_get_header (inode, &tcfsheader, oldkey);
	if (error<0) {
		return error;
	}

	if (!memcmp (tcfsheader.engine, newcipher, TCFS_ENGINE_FIELD_SIZE)) {
		return -TCFS_ERROK;
	}

	uk=tcfs_key_lookup (current->uid, newcipher);
	if (!uk) {
		return -TCFS_ERRACCES;
	}

	newkey=uk->key;

	if (S_ISDIR(inode->i_mode)) {
		error=tcfs_reencrypt_dir (filp, newcipher, &tcfsheader,
					oldkey, newkey);
	} else {
		error=tcfs_reencrypt_file (inode, newcipher, &tcfsheader);
	}

	if (error<0) {
		return error;
	}

	memcpy (tcfsheader.engine, newcipher, TCFS_ENGINE_FIELD_SIZE);
	memcpy (inode->u.tcfs_i.tcfsheader.engine, newcipher,
		TCFS_ENGINE_FIELD_SIZE);

	error=tcfs_put_header (filp, NULL, &tcfsheader, newkey);
	if (error<0) {
		return error;
	}

	return -TCFS_ERROK;
}

static int
tcfs_size_changed (struct dentry *dentry, off_t newsize, void *key)
{
	int error=-TCFS_ERROK;
	struct tcfs_header tcfsheader;

	error=tcfs_get_header (dentry->d_inode, &tcfsheader, key);
	if (error<0) {
		return error;
	}

	tcfsheader.size=newsize;
	dentry->d_inode->u.tcfs_i.tcfsheader.size=newsize;

	error=tcfs_put_header (NULL, dentry, &tcfsheader, key);

	if (error<0) {
		return error;
	}

	return -TCFS_ERROK;
}

static int
old_tcfs_size_changed (struct inode *inode, off_t newsize, void *key)
{
	int error=-TCFS_ERROK;
	char *buffer=NULL, *hash=NULL;
	struct tcfs_fattr fattr;
	struct tcfs_header *tcfsheader=NULL;

	buffer=kmalloc(TCFS_HEADER_SIZE+TCFS_HASH_SIZE, GFP_KERNEL);
	hash=kmalloc(TCFS_HASH_SIZE, GFP_KERNEL);
	if (!buffer && !hash) {
		return -ENOMEM;
	}

	memset (buffer, 0, TCFS_HEADER_SIZE+TCFS_HASH_SIZE);

	error=tcfs_proc_read (TCFS_SERVER(inode), TCFS_INODE_FH(inode),
		0, 0, TCFS_HEADER_SIZE+TCFS_HASH_SIZE, buffer, &fattr);
	if (error<0) {
		kfree (buffer);
		kfree (hash);

		return error;
	}

	error=tcfs_check_hash (buffer, NULL, buffer+TCFS_HEADER_SIZE,
		TCFS_HEADER_SIZE);
	if (error<0) {
		kfree (buffer);
		kfree (hash);

		return -TCFS_ERRACCES;
	}

	tcfsheader=(struct tcfs_header *)buffer;

	tcfsheader->size=newsize;
	inode->u.tcfs_i.tcfsheader.size=newsize;

	tcfs_hash_block (buffer, TCFS_HEADER_SIZE, hash);
	memcpy (buffer+TCFS_HEADER_SIZE, hash, TCFS_HASH_SIZE);
	error=tcfs_proc_write (TCFS_SERVER(inode), TCFS_INODE_FH(inode),
		0, 0, TCFS_HEADER_SIZE+TCFS_HASH_SIZE, buffer, &fattr);

	kfree (buffer);
	kfree (hash);

	if (error<0) {
		return error;
	}

	return -TCFS_ERROK;
}

/*
 * This is called from the tcfs_ioctl (to set the flags or the ciphername) and
 * from the write function (for the file size).
 * It will encrypt/decrypt a file or a dir (i.e. the filenames) according to
 * the flags, if called from tcfs_ioctl.
 */
int
tcfs_modify_header (struct file *filp, struct dentry *dentry, unsigned int what, void *val, void *key)
{
	int error=-TCFS_ERROK;
	struct inode *inode=NULL;

	if (dentry) {
		inode=dentry->d_inode;
	}

	if (filp) {
		inode=filp->f_dentry->d_inode;
	}

	if (!key) {
		key=tcfs_get_key (inode);

		if (what==TCFS_FIELD_FLAGS) {
			struct tcfs_flags *tcfsflags;

			tcfsflags=(struct tcfs_flags *)val;

			if (key==TCFS_NO_SECURE &&
					tcfsflags->flags&TCFS_SECURE_FL) {
				struct user_key *uk=NULL;

				uk=tcfs_key_lookup (current->uid,
							tcfsflags->engine);

				if (!uk) {
					return -TCFS_ERRACCES;
				}

				key=uk->key;
			}

			if (key==TCFS_NO_SECURE &&
					 tcfsflags->flags&TCFS_SHARED_FL) {
				struct group_key *gk=NULL;

				gk=tcfs_groupkey_lookup (inode->i_gid,
					tcfsflags->engine);
				if (!gk || !gk->key) {
					return -TCFS_ERRACCES;
				}

				if (!gk->key) {
					return -TCFS_ERRACCES;
				}

				key=gk->key;
			}
		} 
	}

	/* Ok, if key is still NULL then no access is permitted */

	if (!key) {
		return -TCFS_ERRACCES;
	}
	
	if (what==TCFS_FIELD_ENGINE && key==TCFS_NO_SECURE) {
		return -TCFS_ERRGEN;
	}

	if (what==TCFS_FIELD_SIZE && key==TCFS_NO_SECURE) {
		return -TCFS_ERRGEN;
	}

	switch (what) {
		case TCFS_FIELD_FLAGS:
			error=tcfs_flags_changed (filp,
					 (struct tcfs_flags *)val, key);
			break;

		case TCFS_FIELD_ENGINE:
			error=tcfs_engine_changed (filp, (char *)val, key);
			break;

		case TCFS_FIELD_SIZE:
			error=tcfs_size_changed (dentry, *(off_t *)val, key);
			break;
	}

	if (error<0) {
		return error;
	}

	return -TCFS_ERROK;
}

/* We assume TCFS_HEADER_SIZE <= TCFS_BLOCK_SIZE */
unsigned long
tcfs_new_offset(unsigned long old_offset, unsigned long *blockstart, unsigned long *blockoff)
{
	unsigned long blocks=0;
	unsigned long result=0;

	blocks=tcfs_ldiv (old_offset, TCFS_BLOCK_SIZE);
	*blockstart=TCFS_HEADER_SIZE+TCFS_HASH_SIZE+blocks*TCFS_HBLOCK_SIZE;
	result=TCFS_HEADER_SIZE+TCFS_HASH_SIZE+old_offset+blocks*TCFS_HASH_SIZE;
	*blockoff=result-*blockstart;

	return result;
}
