#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#define _GNU_SOURCE
#include <getopt.h>
#include <string.h>
#include <tcfslib.h>

#include "tcfsutils.h"

#define printerr(format, args...)	fprintf (stderr, format, ## args)

#define USAGE \
"TCFS Utilities: V%d.%d.%d\n\
Set/get flags/cipher of TCFS files.\n\
Usage: tcfsflags [options]... [files]...\n\
Options:\n\
\t-e\t\t\tEncrypt <files>\n\
\t-d\t\t\tDecrypt <files>\n\
\t-g <gid>\t\tShare/Unshare <files> with group <gid>\n\
\t-c <cipher>\t\tUse <cipher> module on <files>\n\
\t-v, --verbose\t\tVerbose mode\n\
\t-V, --version\t\tShow version\n\
\t-h, --help\t\tDisplay this informations.\n"

static struct option long_options[]={
	{ "encrypt", no_argument, NULL, 'e' },
	{ "decrypt", no_argument, NULL, 'd' },
	{ "group", required_argument, NULL, 'g' },
	{ "cipher", required_argument, NULL, 'c' },
	{ "version", no_argument, NULL, 'V' },
	{ "verbose", no_argument, NULL, 'v' },
	{ "info", no_argument, NULL, 'i' },
	{ "help", no_argument, NULL, 'h' },
	{ 0, 0, 0, 0 }
};	

static char *optstring="edg:c:hvVi";
static int verbose=0;

void
tcfs_print_info (char *filename, struct tcfs_header *tcfsheader)
{
	char encrypted='-', shared='-';
	char *engine="clean";
	
	if (tcfsheader)  {
		engine=tcfsheader->engine;

		if (tcfsheader->flags&TCFS_SECURE_FL)
			encrypted='e';
		if (tcfsheader->flags&TCFS_SHARED_FL)
			shared='s';
	}

	printf ("%c%c\t%s\t\t%s\n", encrypted, shared, engine, filename);
}

int main (int argc, char **argv)
{
	int option_index=0;
	int c;
	int encrypt=0, decrypt=0, group=-1, getinfo=0, res=0, version=0;
	char *cipher=NULL;

	if (argc<2) {
		printerr ("Use --help for help\n");
	}

	while ((c=getopt_long (argc, argv, optstring, long_options, &option_index))!=-1) {
		switch (c) {
			case 'v':
				verbose=1;
				break;
			case 'i':
				getinfo=1;
				break;
			case 'e':
				if (decrypt) {
					printerr ("You specified both -e and -d.\n");
					return -1;
				}
				encrypt=1;
				break;
			case 'd':
				if (encrypt) {
					printerr ("You specified both -e and -d.\n");
					return -1;
				}
				decrypt=1;
				break;
			case 'g':
				group=atoi (optarg);
				break;
			case 'c':
				if (decrypt) {
					printerr ("You specified both -d and -c.\n");
					return -1;
				}
				cipher=calloc (TCFS_ENGINE_FIELD_SIZE,
					sizeof (char));
				memset (cipher, 0, TCFS_ENGINE_FIELD_SIZE);
				strcpy (cipher, optarg);
				break;
			case 'h':
				printf (USAGE, MAJOR, MINOR, RELEASE);
				return -1;
			case 'V':
				version=1;
				break;
		}
	};

	if (version) {
		printf ("tcfsflags V%d.%d.%d\n", MAJOR, MINOR, RELEASE);
	}

	if (optind == argc) {
		printerr ("%s: <filename>(s) required\n", argv[0]);
		return 0;
	}

	if (group>-1 && !encrypt && !decrypt) {
		printerr ("%s: -g option cannot be used without -e or -d\n",
				argv[0]);
		return 0;
	}

	if (verbose && getinfo) {
		printf ("Flags\tCipher\t\tFilename\n");
	}
	
	while ( optind < argc ) {
		char *filename=argv[optind++];

		if (encrypt && !cipher) {
			printerr ("%s: a <cipher_name> must be specified\n", argv[0]);
			return 0;
		}

		if (encrypt) {
			res=tcfs_encrypt_file (filename, cipher, group);
		}
	
		if (decrypt) {
			res=tcfs_decrypt_file (filename);
		}

		if (cipher && !encrypt) {
			res=tcfs_change_cipher (filename, cipher);
		}

		if (getinfo) {
			struct tcfs_header *tcfsheader;

			tcfsheader=tcfs_get_header (filename);
			tcfs_print_info (filename, tcfsheader);

			free (tcfsheader);
		}
	}

	if (res!=0) {
		printf("%s\n",tcfs_strerror(res));
	} 

	return 0;
}
